@@ -1,4 +1,5 @@
-from datetime import datetime
+import asyncio
+from datetime import datetime, timezone
 from logging import Logger
 
 import jinja2
@@ -7,6 +8,7 @@ from unique_toolkit.agentic.debug_info_manager.debug_info_manager import (
 )
 from unique_toolkit.agentic.evaluation.evaluation_manager import EvaluationManager
 from unique_toolkit.agentic.history_manager.history_manager import HistoryManager
+from unique_toolkit.agentic.message_log_manager.service import MessageStepLogger
 from unique_toolkit.agentic.postprocessor.postprocessor_manager import (
     PostprocessorManager,
 )
@@ -14,6 +16,7 @@ from unique_toolkit.agentic.reference_manager.reference_manager import Reference
 from unique_toolkit.agentic.thinking_manager.thinking_manager import ThinkingManager
 from unique_toolkit.agentic.tools.tool_manager import (
     ResponsesApiToolManager,
+    SafeTaskExecutor,
     ToolManager,
 )
 from unique_toolkit.app.schemas import ChatEvent, McpServer
@@ -59,6 +62,7 @@ class UniqueAI:
         history_manager: HistoryManager,
         evaluation_manager: EvaluationManager,
         postprocessor_manager: PostprocessorManager,
+        message_step_logger: MessageStepLogger,
         mcp_servers: list[McpServer],
     ):
         self._logger = logger
@@ -80,6 +84,7 @@ class UniqueAI:
         self._mcp_servers = mcp_servers
         self._streaming_handler = streaming_handler
 
+        self._message_step_logger = message_step_logger
         # Helper variable to support control loop
         self._tool_took_control = False
 
@@ -133,9 +138,7 @@ class UniqueAI:
             self.start_text = self._thinking_manager.update_start_text(
                 self.start_text, loop_response
             )
-            await self._create_new_assistant_message_if_loop_response_contains_content(
-                loop_response
-            )
+        await self._update_debug_info_if_tool_took_control()
 
         # Only set completed_at if no tool took control. Tools that take control will set the message state to completed themselves.
         await self._chat_service.modify_assistant_message_async(
@@ -229,6 +232,9 @@ class UniqueAI:
             self._logger.debug(
                 "Tools were called we process them and do not exit the loop"
             )
+            await self._create_new_assistant_message_if_loop_response_contains_content(
+                loop_response
+            )
 
             return await self._handle_tool_calls(loop_response)
 
@@ -259,12 +265,17 @@ class UniqueAI:
             for prompts in self._tool_manager.get_tool_prompts()
         ]
 
-        used_tools = [m.name for m in self._tool_manager.get_tools()]
+        used_tools = [t.name for t in self._history_manager.get_tool_calls()]
+        sub_agent_calls = self._tool_manager.filter_tool_calls(
+            self._history_manager.get_tool_calls(), ["subagent"]
+        )
 
         mcp_server_user_prompts = [
             mcp_server.user_prompt for mcp_server in self._mcp_servers
         ]
 
+        user_metadata = self._get_filtered_user_metadata()
+
         tool_descriptions = self._tool_manager.get_tool_prompts()
 
         query = self._event.payload.user_message.text
@@ -272,6 +283,7 @@ class UniqueAI:
         if (
             self._config.agent.experimental.sub_agents_config.referencing_config
             is not None
+            and len(sub_agent_calls) > 0
         ):
             use_sub_agent_references = True
             sub_agent_referencing_instructions = self._config.agent.experimental.sub_agents_config.referencing_config.referencing_instructions_for_user_prompt
@@ -287,16 +299,18 @@ class UniqueAI:
             tool_descriptions_with_user_prompts=tool_descriptions_with_user_prompts,
             use_sub_agent_references=use_sub_agent_references,
             sub_agent_referencing_instructions=sub_agent_referencing_instructions,
+            user_metadata=user_metadata,
         )
         return user_msg
 
-    async def _render_system_prompt(
-        self,
-    ) -> str:
+    async def _render_system_prompt(self) -> str:
         # TODO: Collect tool information here and adapt to system prompt
         tool_descriptions = self._tool_manager.get_tool_prompts()
 
-        used_tools = [m.name for m in self._tool_manager.get_tools()]
+        used_tools = [t.name for t in self._history_manager.get_tool_calls()]
+        sub_agent_calls = self._tool_manager.filter_tool_calls(
+            self._history_manager.get_tool_calls(), ["subagent"]
+        )
 
         system_prompt_template = jinja2.Template(
             self._config.agent.prompt_config.system_prompt_template
@@ -304,6 +318,8 @@ class UniqueAI:
 
         date_string = datetime.now().strftime("%A %B %d, %Y")
 
+        user_metadata = self._get_filtered_user_metadata()
+
         mcp_server_system_prompts = [
             mcp_server.system_prompt for mcp_server in self._mcp_servers
         ]
@@ -311,6 +327,7 @@ class UniqueAI:
         if (
             self._config.agent.experimental.sub_agents_config.referencing_config
             is not None
+            and len(sub_agent_calls) > 0
         ):
             use_sub_agent_references = True
             sub_agent_referencing_instructions = self._config.agent.experimental.sub_agents_config.referencing_config.referencing_instructions_for_system_prompt
@@ -318,6 +335,14 @@ class UniqueAI:
             use_sub_agent_references = False
             sub_agent_referencing_instructions = None
 
+        uploaded_documents = self._content_service.get_documents_uploaded_to_chat()
+        uploaded_documents_expired = [
+            doc
+            for doc in uploaded_documents
+            if doc.expired_at is not None
+            and doc.expired_at <= datetime.now(timezone.utc)
+        ]
+
         system_message = system_prompt_template.render(
             model_info=self._config.space.language_model.model_dump(mode="json"),
             date_string=date_string,
@@ -331,6 +356,8 @@ class UniqueAI:
             mcp_server_system_prompts=mcp_server_system_prompts,
             use_sub_agent_references=use_sub_agent_references,
             sub_agent_referencing_instructions=sub_agent_referencing_instructions,
+            user_metadata=user_metadata,
+            uploaded_documents_expired=uploaded_documents_expired,
         )
         return system_message
 
@@ -338,20 +365,67 @@ class UniqueAI:
         self, loop_response: LanguageModelStreamResponse
     ) -> bool:
         """Handle the case where no tool calls are returned."""
-        selected_evaluation_names = self._tool_manager.get_evaluation_check_list()
-        evaluation_results = await self._evaluation_manager.run_evaluations(
-            selected_evaluation_names, loop_response, self._latest_assistant_id
+        task_executor = SafeTaskExecutor(
+            logger=self._logger,
         )
 
-        await self._postprocessor_manager.run_postprocessors(loop_response)
+        selected_evaluation_names = self._tool_manager.get_evaluation_check_list()
+        evaluation_results = task_executor.execute_async(
+            self._evaluation_manager.run_evaluations,
+            selected_evaluation_names,
+            loop_response,
+            self._latest_assistant_id,
+        )
 
-        if not all(result.is_positive for result in evaluation_results):
+        postprocessor_result = task_executor.execute_async(
+            self._postprocessor_manager.run_postprocessors,
+            loop_response.model_copy(deep=True),
+        )
+
+        _, evaluation_results = await asyncio.gather(
+            postprocessor_result,
+            evaluation_results,
+        )
+
+        if evaluation_results.success and not all(
+            result.is_positive for result in evaluation_results.unpack()
+        ):
             self._logger.warning(
                 "we should add here the retry counter add an instruction and retry the loop for now we just exit the loop"
             )  # TODO: add retry counter and instruction
 
         return True
 
+    def _log_tool_calls(self, tool_calls: list) -> None:
+        # Create dictionary mapping tool names to display names for efficient lookup
+        all_tools_dict: dict[str, str] = {
+            tool.name: tool.display_name()
+            for tool in self._tool_manager.available_tools
+        }
+
+        # Tool names that should not be logged in the message steps
+        tool_names_not_to_log = ["DeepResearch"]
+
+        tool_string: str = ""
+        used_tools: dict[str, int] = {}
+        for tool_call in tool_calls:
+            self._history_manager.add_tool_call(tool_call)
+            if tool_call.name in all_tools_dict:
+                used_tools[tool_call.name] = used_tools.get(tool_call.name, 0) + 1
+
+        for tool_name, count in used_tools.items():
+            if tool_name in tool_names_not_to_log:
+                continue
+            display_name = all_tools_dict[tool_name] or tool_name
+            tool_string += (
+                f"\n• {display_name} ({count}x)" if count > 1 else f"\n• {display_name}"
+            )
+
+        if tool_string:
+            self._message_step_logger.create_message_log_entry(
+                text=f"**Triggered Tool Calls:**\n {tool_string}", references=[]
+            )
+
     async def _handle_tool_calls(
         self, loop_response: LanguageModelStreamResponse
     ) -> bool:
@@ -360,9 +434,17 @@ class UniqueAI:
 
         tool_calls = loop_response.tool_calls or []
 
+        # Filter tool calls
+        tool_calls = self._tool_manager.filter_duplicate_tool_calls(tool_calls)
+        tool_calls = self._tool_manager.filter_tool_calls_by_max_tool_calls_allowed(
+            tool_calls
+        )
+
         # Append function calls to history
         self._history_manager._append_tool_calls_to_history(tool_calls)
 
+        # Log tool calls
+        self._log_tool_calls(tool_calls)
         # Execute tool calls
         tool_call_responses = await self._tool_manager.execute_selected_tools(
             tool_calls
@@ -416,6 +498,52 @@ class UniqueAI:
             )
         )
 
+    def _get_filtered_user_metadata(self) -> dict[str, str]:
+        """
+        Filter user metadata to only include keys specified in the agent's prompt config.
+
+        Returns:
+            Dictionary containing only the metadata keys that are configured to be included.
+        """
+        user_metadata = {}
+        if (
+            self._config.agent.prompt_config.user_metadata
+            and self._event.payload.user_metadata is not None
+        ):
+            # Filter metadata to only include selected keys
+            user_metadata = {
+                k: str(v)
+                for k, v in self._event.payload.user_metadata.items()
+                if k in self._config.agent.prompt_config.user_metadata
+            }
+        return user_metadata
+
+    async def _update_debug_info_if_tool_took_control(self) -> None:
+        """
+        Update debug info when a tool takes control of the conversation.
+        DeepResearch is excluded as it handles debug info directly since it calls
+        the orchestrator multiple times.
+        """
+        if not self._tool_took_control:
+            return
+
+        tool_names = [tool["name"] for tool in self._debug_info_manager.get()["tools"]]
+        if "DeepResearch" in tool_names:
+            return
+
+        debug_info_event = {
+            "assistant": {
+                "id": self._event.payload.assistant_id,
+                "name": self._event.payload.name,
+            },
+            "chosenModule": self._event.payload.name,
+            "userMetadata": self._event.payload.user_metadata,
+            "toolParameters": self._event.payload.tool_parameters,
+            **self._debug_info_manager.get(),
+        }
+        
+        await self._chat_service.update_debug_info_async(debug_info=debug_info_event)
+
 
 class UniqueAIResponsesApi(UniqueAI):
     def __init__(
@@ -433,6 +561,7 @@ class UniqueAIResponsesApi(UniqueAI):
         history_manager: HistoryManager,
         evaluation_manager: EvaluationManager,
         postprocessor_manager: PostprocessorManager,
+        message_step_logger: MessageStepLogger,
         mcp_servers: list[McpServer],
     ) -> None:
         super().__init__(
@@ -449,5 +578,6 @@ class UniqueAIResponsesApi(UniqueAI):
             history_manager=history_manager,
             evaluation_manager=evaluation_manager,
             postprocessor_manager=postprocessor_manager,
+            message_step_logger=message_step_logger,
             mcp_servers=mcp_servers,
         )
