@@ -1,8 +1,7 @@
-import os
+from datetime import datetime, timezone
 from logging import Logger
-from typing import NamedTuple
+from typing import NamedTuple, cast
 
-from openai import AsyncOpenAI
 from unique_follow_up_questions.follow_up_postprocessor import (
     FollowUpPostprocessor,
 )
@@ -30,14 +29,13 @@ from unique_toolkit.agentic.history_manager.history_manager import (
     HistoryManager,
     HistoryManagerConfig,
 )
+from unique_toolkit.agentic.message_log_manager.service import MessageStepLogger
 from unique_toolkit.agentic.postprocessor.postprocessor_manager import (
-    Postprocessor,
     PostprocessorManager,
 )
 from unique_toolkit.agentic.reference_manager.reference_manager import ReferenceManager
 from unique_toolkit.agentic.responses_api import (
     DisplayCodeInterpreterFilesPostProcessor,
-    DisplayCodeInterpreterFilesPostProcessorConfig,
     ShowExecutedCodePostprocessor,
 )
 from unique_toolkit.agentic.thinking_manager.thinking_manager import (
@@ -47,11 +45,17 @@ from unique_toolkit.agentic.thinking_manager.thinking_manager import (
 from unique_toolkit.agentic.tools.a2a import (
     A2AManager,
     ExtendedSubAgentToolConfig,
+    SubAgentDisplaySpec,
     SubAgentEvaluationService,
-    SubAgentResponsesPostprocessor,
+    SubAgentEvaluationSpec,
+    SubAgentReferencesPostprocessor,
+    SubAgentResponsesDisplayPostprocessor,
+    SubAgentResponsesPostprocessorConfig,
+    SubAgentResponseWatcher,
 )
 from unique_toolkit.agentic.tools.config import ToolBuildConfig
 from unique_toolkit.agentic.tools.mcp.manager import MCPManager
+from unique_toolkit.agentic.tools.openai_builtin.base import OpenAIBuiltInToolName
 from unique_toolkit.agentic.tools.tool_manager import (
     OpenAIBuiltInToolManager,
     ResponsesApiToolManager,
@@ -103,13 +107,15 @@ class _CommonComponents(NamedTuple):
     reference_manager: ReferenceManager
     history_manager: HistoryManager
     evaluation_manager: EvaluationManager
+    postprocessor_manager: PostprocessorManager
+    message_step_logger: MessageStepLogger
+    response_watcher: SubAgentResponseWatcher
     # Tool Manager Components
     tool_progress_reporter: ToolProgressReporter
     tool_manager_config: ToolManagerConfig
     mcp_manager: MCPManager
     a2a_manager: A2AManager
     mcp_servers: list[McpServer]
-    postprocessors: list[Postprocessor]
 
 
 def _build_common(
@@ -123,7 +129,12 @@ def _build_common(
 
     uploaded_documents = content_service.get_documents_uploaded_to_chat()
 
-    tool_progress_reporter = ToolProgressReporter(chat_service=chat_service)
+    response_watcher = SubAgentResponseWatcher()
+
+    tool_progress_reporter = ToolProgressReporter(
+        chat_service=chat_service,
+        config=config.agent.services.tool_progress_reporter_config,
+    )
     thinking_manager_config = ThinkingManagerConfig(
         thinking_steps_display=config.agent.experimental.thinking_steps_display
     )
@@ -137,9 +148,7 @@ def _build_common(
     reference_manager = ReferenceManager()
 
     history_manager_config = HistoryManagerConfig(
-        experimental_features=history_manager_module.ExperimentalFeatures(
-            full_sources_serialize_dump=False,
-        ),
+        experimental_features=history_manager_module.ExperimentalFeatures(),
         percent_of_max_tokens_for_history=config.agent.input_token_distribution.percent_for_history,
         language_model=config.space.language_model,
         uploaded_content_config=config.agent.services.uploaded_content_config,
@@ -170,16 +179,21 @@ def _build_common(
     a2a_manager = A2AManager(
         logger=logger,
         tool_progress_reporter=tool_progress_reporter,
+        response_watcher=response_watcher,
     )
+
     tool_manager_config = ToolManagerConfig(
         tools=config.space.tools,
         max_tool_calls=config.agent.experimental.loop_configuration.max_tool_calls_per_iteration,
     )
 
-    postprocessors = []
+    postprocessor_manager = PostprocessorManager(
+        logger=logger,
+        chat_service=chat_service,
+    )
 
-    if config.agent.services.stock_ticker_config:
-        postprocessors.append(
+    if config.agent.services.stock_ticker_config is not None:
+        postprocessor_manager.add_postprocessor(
             StockTickerPostprocessor(
                 config=config.agent.services.stock_ticker_config,
                 event=event,
@@ -190,7 +204,8 @@ def _build_common(
         config.agent.services.follow_up_questions_config
         and config.agent.services.follow_up_questions_config.number_of_questions > 0
     ):
-        postprocessors.append(
+        # Should run last to make sure the follow up questions are displayed last.
+        postprocessor_manager.set_last_postprocessor(
             FollowUpPostprocessor(
                 logger=logger,
                 config=config.agent.services.follow_up_questions_config,
@@ -213,42 +228,12 @@ def _build_common(
         a2a_manager=a2a_manager,
         tool_manager_config=tool_manager_config,
         mcp_servers=event.payload.mcp_servers,
-        postprocessors=postprocessors,
+        postprocessor_manager=postprocessor_manager,
+        response_watcher=response_watcher,
+        message_step_logger=MessageStepLogger(chat_service),
     )
 
 
-def _prepare_base_url(url: str, use_v1: bool) -> str:
-    url = url.rstrip("/") + "/openai"
-
-    if use_v1:
-        url += "/v1"
-
-    return url
-
-
-def _get_openai_client_from_env(
-    config: UniqueAIConfig, use_v1: bool = False
-) -> AsyncOpenAI:
-    use_direct_azure_client = (
-        config.agent.experimental.responses_api_config.use_direct_azure_client
-    )
-    api_key_env_var = config.agent.experimental.responses_api_config.direct_azure_client_api_key_env_var
-    api_base_env_var = config.agent.experimental.responses_api_config.direct_azure_client_api_base_env_var
-
-    if use_direct_azure_client:
-        # TODO: (for testing only), remove when v1 endpoint is working
-        return AsyncOpenAI(
-            api_key=os.environ[api_key_env_var],
-            base_url=_prepare_base_url(os.environ[api_base_env_var], use_v1=use_v1),
-        )
-    else:
-        return get_async_openai_client().copy(
-            default_headers={
-                "x-model": config.space.language_model.name
-            }  # Backend requires a model name
-        )
-
-
 async def _build_responses(
     event: ChatEvent,
     logger: Logger,
@@ -256,17 +241,66 @@ async def _build_responses(
     common_components: _CommonComponents,
     debug_info_manager: DebugInfoManager,
 ) -> UniqueAIResponsesApi:
-    client = _get_openai_client_from_env(config, use_v1=True)
-    builtin_tool_manager = OpenAIBuiltInToolManager(
+    client = get_async_openai_client().copy(
+        default_headers={
+            "x-model": config.space.language_model.name,
+            "x-user-id": event.user_id,
+            "x-company-id": event.company_id,
+            "x-assistant-id": event.payload.assistant_id,
+            "x-chat-id": event.payload.chat_id,
+        }
+    )
+
+    assert config.agent.experimental.responses_api_config is not None
+
+    code_interpreter_config = (
+        config.agent.experimental.responses_api_config.code_interpreter
+    )
+    postprocessor_manager = common_components.postprocessor_manager
+    tool_names = [tool.name for tool in config.space.tools]
+
+    if code_interpreter_config is not None:
+        if OpenAIBuiltInToolName.CODE_INTERPRETER not in tool_names:
+            logger.info("Automatically adding code interpreter to the tools")
+            config = config.model_copy(deep=True)
+            config.space.tools.append(
+                ToolBuildConfig(
+                    name=OpenAIBuiltInToolName.CODE_INTERPRETER,
+                    configuration=code_interpreter_config.tool_config,
+                )
+            )
+            common_components.tool_manager_config.tools = config.space.tools
+
+        if code_interpreter_config.executed_code_display_config is not None:
+            postprocessor_manager.add_postprocessor(
+                ShowExecutedCodePostprocessor(
+                    config=code_interpreter_config.executed_code_display_config
+                )
+            )
+
+        postprocessor_manager.add_postprocessor(
+            DisplayCodeInterpreterFilesPostProcessor(
+                client=client,
+                content_service=common_components.content_service,
+                config=code_interpreter_config.generated_files_config,
+                user_id=event.user_id,
+                company_id=event.company_id,
+                chat_id=event.payload.chat_id,
+                chat_service=common_components.chat_service,
+            )
+        )
+
+    builtin_tool_manager = await OpenAIBuiltInToolManager.build_manager(
         uploaded_files=common_components.uploaded_documents,
-        chat_id=event.payload.chat_id,
         content_service=common_components.content_service,
         user_id=event.user_id,
         company_id=event.company_id,
+        chat_id=event.payload.chat_id,
         client=client,
+        tool_configs=config.space.tools,
     )
 
-    tool_manager = await ResponsesApiToolManager.build_manager(
+    tool_manager = ResponsesApiToolManager(
         logger=logger,
         config=common_components.tool_manager_config,
         event=event,
@@ -276,32 +310,7 @@ async def _build_responses(
         builtin_tool_manager=builtin_tool_manager,
     )
 
-    postprocessor_manager = PostprocessorManager(
-        logger=logger,
-        chat_service=common_components.chat_service,
-    )
-    for postprocessor in common_components.postprocessors:
-        postprocessor_manager.add_postprocessor(postprocessor)
-
-    if (
-        config.agent.experimental.responses_api_config.code_interpreter_display_config
-        is not None
-    ):
-        postprocessor_manager.add_postprocessor(
-            ShowExecutedCodePostprocessor(
-                config=config.agent.experimental.responses_api_config.code_interpreter_display_config
-            )
-        )
-
-    postprocessor_manager.add_postprocessor(
-        DisplayCodeInterpreterFilesPostProcessor(
-            client=client,
-            content_service=common_components.content_service,
-            config=DisplayCodeInterpreterFilesPostProcessorConfig(
-                upload_scope_id=config.agent.experimental.responses_api_config.generated_files_scope_id,
-            ),
-        )
-    )
+    postprocessor_manager = common_components.postprocessor_manager
 
     class ResponsesStreamingHandler(ResponsesSupportCompleteWithReferences):
         def complete_with_references(self, *args, **kwargs):
@@ -319,15 +328,15 @@ async def _build_responses(
     _add_sub_agents_postprocessor(
         postprocessor_manager=postprocessor_manager,
         tool_manager=tool_manager,
-        user_id=event.user_id,
-        company_id=event.company_id,
-        chat_id=event.payload.chat_id,
+        config=config,
+        response_watcher=common_components.response_watcher,
     )
     _add_sub_agents_evaluation(
         evaluation_manager=common_components.evaluation_manager,
         tool_manager=tool_manager,
         config=config,
         event=event,
+        response_watcher=common_components.response_watcher,
     )
 
     return UniqueAIResponsesApi(
@@ -344,6 +353,7 @@ async def _build_responses(
         evaluation_manager=common_components.evaluation_manager,
         postprocessor_manager=postprocessor_manager,
         debug_info_manager=debug_info_manager,
+        message_step_logger=common_components.message_step_logger,
         mcp_servers=event.payload.mcp_servers,
     )
 
@@ -355,17 +365,40 @@ def _build_completions(
     common_components: _CommonComponents,
     debug_info_manager: DebugInfoManager,
 ) -> UniqueAI:
-    if len(common_components.uploaded_documents) > 0:
+    # Uploaded content behavior is always to force uploaded search tool:
+    # 1. Add it to forced tools if there are tool choices.
+    # 2. Simply force it if there are no tool choices.
+    # 3. Not available if not uploaded documents.
+    now = datetime.now(timezone.utc)
+    UPLOADED_DOCUMENTS_VALID = [
+        doc
+        for doc in common_components.uploaded_documents
+        if doc.expired_at is None or doc.expired_at > now
+    ]
+    UPLOADED_DOCUMENTS_EXPIRED = [
+        doc
+        for doc in common_components.uploaded_documents
+        if doc.expired_at is not None and doc.expired_at <= now
+    ]
+    TOOL_CHOICES = len(event.payload.tool_choices) > 0
+
+    if UPLOADED_DOCUMENTS_EXPIRED:
         logger.info(
-            f"Adding UploadedSearchTool with {len(common_components.uploaded_documents)} documents"
+            f"Number of expired uploaded documents: {len(UPLOADED_DOCUMENTS_EXPIRED)}"
         )
-        config.space.tools.append(
+
+    if UPLOADED_DOCUMENTS_VALID:
+        logger.info(
+            f"Number of valid uploaded documents: {len(UPLOADED_DOCUMENTS_VALID)}"
+        )
+        common_components.tool_manager_config.tools.append(
             ToolBuildConfig(
                 name=UploadedSearchTool.name,
                 display_name=UploadedSearchTool.name,
                 configuration=UploadedSearchConfig(),
-            ),
+            )
         )
+    if TOOL_CHOICES and UPLOADED_DOCUMENTS_VALID:
         event.payload.tool_choices.append(str(UploadedSearchTool.name))
 
     tool_manager = ToolManager(
@@ -376,26 +409,23 @@ def _build_completions(
         mcp_manager=common_components.mcp_manager,
         a2a_manager=common_components.a2a_manager,
     )
+    if not TOOL_CHOICES and UPLOADED_DOCUMENTS_VALID:
+        tool_manager.add_forced_tool(UploadedSearchTool.name)
 
-    postprocessor_manager = PostprocessorManager(
-        logger=logger,
-        chat_service=common_components.chat_service,
-    )
-    for postprocessor in common_components.postprocessors:
-        postprocessor_manager.add_postprocessor(postprocessor)
+    postprocessor_manager = common_components.postprocessor_manager
 
     _add_sub_agents_postprocessor(
         postprocessor_manager=postprocessor_manager,
         tool_manager=tool_manager,
-        user_id=event.user_id,
-        company_id=event.company_id,
-        chat_id=event.payload.chat_id,
+        config=config,
+        response_watcher=common_components.response_watcher,
     )
     _add_sub_agents_evaluation(
         evaluation_manager=common_components.evaluation_manager,
         tool_manager=tool_manager,
         config=config,
         event=event,
+        response_watcher=common_components.response_watcher,
     )
 
     return UniqueAI(
@@ -413,30 +443,44 @@ def _build_completions(
         postprocessor_manager=postprocessor_manager,
         debug_info_manager=debug_info_manager,
         mcp_servers=event.payload.mcp_servers,
+        message_step_logger=common_components.message_step_logger,
     )
 
 
 def _add_sub_agents_postprocessor(
     postprocessor_manager: PostprocessorManager,
     tool_manager: ToolManager | ResponsesApiToolManager,
-    user_id: str,
-    company_id: str,
-    chat_id: str,
+    config: UniqueAIConfig,
+    response_watcher: SubAgentResponseWatcher,
 ) -> None:
     sub_agents = tool_manager.sub_agents
     if len(sub_agents) > 0:
-        sub_agent_responses_postprocessor = SubAgentResponsesPostprocessor(
-            user_id=user_id,
-            main_agent_chat_id=chat_id,
-            company_id=company_id,
+        display_config = SubAgentResponsesPostprocessorConfig(
+            sleep_time_before_update=config.agent.experimental.sub_agents_config.sleep_time_before_update,
         )
-        postprocessor_manager.add_postprocessor(sub_agent_responses_postprocessor)
+        display_specs = []
+        for tool in sub_agents:
+            tool_config = cast(
+                ExtendedSubAgentToolConfig, tool.settings.configuration
+            )  # (BeforeValidator of ToolBuildConfig)
 
-        for tool in tool_manager.sub_agents:
-            assert isinstance(tool.config, ExtendedSubAgentToolConfig)
-            sub_agent_responses_postprocessor.register_sub_agent_tool(
-                tool, tool.config.response_display_config
+            display_specs.append(
+                SubAgentDisplaySpec(
+                    assistant_id=tool_config.assistant_id,
+                    display_name=tool.display_name(),
+                    display_config=tool_config.response_display_config,
+                )
             )
+        reference_postprocessor = SubAgentReferencesPostprocessor(
+            response_watcher=response_watcher,
+        )
+        sub_agent_responses_postprocessor = SubAgentResponsesDisplayPostprocessor(
+            config=display_config,
+            response_watcher=response_watcher,
+            display_specs=display_specs,
+        )
+        postprocessor_manager.add_postprocessor(reference_postprocessor)
+        postprocessor_manager.add_postprocessor(sub_agent_responses_postprocessor)
 
 
 def _add_sub_agents_evaluation(
@@ -444,18 +488,31 @@ def _add_sub_agents_evaluation(
     tool_manager: ToolManager | ResponsesApiToolManager,
     config: UniqueAIConfig,
     event: ChatEvent,
+    response_watcher: SubAgentResponseWatcher,
 ) -> None:
     sub_agents = tool_manager.sub_agents
-    if len(sub_agents) > 0:
-        sub_agent_evaluation = None
-        if config.agent.experimental.sub_agents_config.evaluation_config is not None:
-            sub_agent_evaluation = SubAgentEvaluationService(
-                config=config.agent.experimental.sub_agents_config.evaluation_config,
-                language_model_service=LanguageModelService.from_event(event),
-            )
-            evaluation_manager.add_evaluation(sub_agent_evaluation)
-            for tool in tool_manager.sub_agents:
-                assert isinstance(tool.config, ExtendedSubAgentToolConfig)
-                sub_agent_evaluation.register_sub_agent_tool(
-                    tool, tool.config.evaluation_config
+    if (
+        len(sub_agents) > 0
+        and config.agent.experimental.sub_agents_config.evaluation_config is not None
+    ):
+        evaluation_specs = []
+        for tool in sub_agents:
+            tool_config = cast(
+                ExtendedSubAgentToolConfig, tool.settings.configuration
+            )  # (BeforeValidator of ToolBuildConfig)
+
+            evaluation_specs.append(
+                SubAgentEvaluationSpec(
+                    assistant_id=tool_config.assistant_id,
+                    display_name=tool.display_name(),
+                    config=tool_config.evaluation_config,
                 )
+            )
+
+        sub_agent_evaluation = SubAgentEvaluationService(
+            config=config.agent.experimental.sub_agents_config.evaluation_config,
+            language_model_service=LanguageModelService.from_event(event),
+            evaluation_specs=evaluation_specs,
+            response_watcher=response_watcher,
+        )
+        evaluation_manager.add_evaluation(sub_agent_evaluation)
